import shutil

from typing import Union, Literal

from .maths import chars_len
from .colors import colored, decolored, COLOR_TYPE


def fill_to_len(filler: str, length: int) -> str:
    filler_len = chars_len(filler)
    complete_fill_count = length // filler_len
    remaining_fill_len = length % filler_len
    filled_str = filler * complete_fill_count + filler[:remaining_fill_len]
    return filled_str


def add_fills(
    text: str = "",
    filler: str = "=",
    fill_side: Literal["left", "right", "both"] = "both",
    is_text_colored: bool = False,
    fill_color: COLOR_TYPE = None,
    total_width: int = None,
):
    if not total_width:
        total_width = shutil.get_terminal_size().columns
    if not text:
        filled_str = colored(fill_to_len(filler, total_width), color=fill_color)
        return filled_str

    text = text.strip()
    if is_text_colored:
        text_width = chars_len(decolored(text))
    else:
        text_width = chars_len(text)
    if text_width >= total_width:
        return text

    if fill_side[0].lower() == "b":
        leading_fill_str = (
            fill_to_len(filler, (total_width - text_width) // 2 - 1) + " "
        )
        trailing_fill_str = " " + fill_to_len(
            filler, total_width - text_width - chars_len(leading_fill_str) - 1
        )
    elif fill_side[0].lower() == "l":
        leading_fill_str = fill_to_len(filler, total_width - text_width - 1) + " "
        trailing_fill_str = ""
    elif fill_side[0].lower() == "r":
        leading_fill_str = ""
        trailing_fill_str = " " + fill_to_len(filler, total_width - text_width - 1)
    else:
        raise ValueError("Invalid fill_side")

    if fill_color:
        leading_fill_str = colored(leading_fill_str, color=fill_color)
        trailing_fill_str = colored(trailing_fill_str, color=fill_color)

    filled_str = f"{leading_fill_str}{text}{trailing_fill_str}"
    return filled_str



import functools
import logging

from typing import Literal

from .colors import colored
from .fills import add_fills
from .times import get_now

LOG_METHOD_COLORS = {
    "err": ("error", "red"),
    "erro": ("error", "red"),
    "warn": ("warning", "light_red"),
    "hint": ("info", "light_yellow"),
    "glow": ("info", "black"),
    "note": ("info", "light_magenta"),
    "mesg": ("info", "light_cyan"),
    "file": ("info", "light_blue"),
    "line": ("info", "white"),
    "okay": ("info", "light_green"),
    "success": ("info", "light_green"),
    "fail": ("info", "light_red"),
    "back": ("debug", "light_cyan"),
    "dbug": ("debug", "light_cyan"),
}

LOG_METHOD_BG_COLORS = {
    "glow": "bg_blue",
}


class TCLogstr:
    def __init__(self):
        self.COLORS = {k: v[1] for k, v in LOG_METHOD_COLORS.items()}

    def colored_str(self, msg, method, *args, **kwargs):
        return colored(
            msg,
            color=self.COLORS[method.lower()],
            bg_color=LOG_METHOD_BG_COLORS.get(method, None),
            *args,
            **kwargs,
        )

    def err(self, msg: str = ""):
        return self.colored_str(msg, "err")

    def erro(self, msg: str = ""):
        return self.colored_str(msg, "erro")

    def warn(self, msg: str = ""):
        return self.colored_str(msg, "warn")

    def hint(self, msg: str = ""):
        return self.colored_str(msg, "hint")

    def glow(self, msg: str = ""):
        return self.colored_str(msg, "glow")

    def note(self, msg: str = ""):
        return self.colored_str(msg, "note")

    def mesg(self, msg: str = ""):
        return self.colored_str(msg, "mesg")

    def file(self, msg: str = ""):
        return self.colored_str(msg, "file")

    def line(self, msg: str = ""):
        return self.colored_str(msg, "line")

    def success(self, msg: str = ""):
        return self.colored_str(msg, "success")

    def okay(self, msg: str = ""):
        return self.colored_str(msg, "okay")

    def fail(self, msg: str = ""):
        return self.colored_str(msg, "fail")

    def back(self, msg: str = ""):
        return self.colored_str(msg, "back")

    def dbug(self, msg: str = ""):
        return self.colored_str(msg, "dbug")


logstr = TCLogstr()


class TCLogger(logging.Logger):
    INDENT_METHODS = [
        "indent",
        "set_indent",
        "reset_indent",
        "store_indent",
        "restore_indent",
        "log_indent",
    ]
    LEVEL_METHODS = [
        "set_level",
        "store_level",
        "restore_level",
        "quiet",
        "enter_quiet",
        "exit_quiet",
    ]
    LEVEL_NAMES = {
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG,
    }

    def __init__(
        self,
        name: str = None,
        use_prefix: bool = False,
        use_prefix_ms: bool = False,
        use_prefix_color: bool = False,
        verbose: bool = True,
    ):
        self.name = str(name) if name is not None else "TCLogger"
        self.use_prefix = use_prefix
        self.use_prefix_ms = use_prefix_ms
        self.use_prefix_color = use_prefix_color
        self.verbose = verbose

        super().__init__(self.name)
        self.setLevel(logging.INFO)
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)
        self.addHandler(stream_handler)
        self.log_indent = 0
        self.log_indents = []
        self.log_level = "info"
        self.log_levels = []

    def indent(self, indent=2):
        self.log_indent += indent

    def set_indent(self, indent=2):
        self.log_indent = indent

    def reset_indent(self):
        self.log_indent = 0

    def store_indent(self):
        self.log_indents.append(self.log_indent)

    def restore_indent(self):
        self.log_indent = self.log_indents.pop(-1)

    def set_level(
        self, level: Literal["critical", "error", "warning", "info", "debug"]
    ):
        self.log_level = level
        self.setLevel(self.LEVEL_NAMES[level])
        for handler in self.handlers:
            handler.setLevel(self.LEVEL_NAMES[level])

    def store_level(self):
        self.log_levels.append(self.log_level)

    def restore_level(self):
        self.log_level = self.log_levels.pop(-1)
        self.set_level(self.log_level)

    def quiet(self):
        self.set_level("error")

    def enter_quiet(self, quiet=False):
        if quiet:
            self.store_level()
            self.quiet()

    def exit_quiet(self, quiet=False):
        if quiet:
            self.restore_level()

    def get_prefix_str(self, method: str) -> str:
        """Generate prefix string with timestamp, log level, and logger name."""
        now = get_now()
        if self.use_prefix_ms:
            time_str = (
                now.strftime("%Y-%m-%d %H:%M:%S") + f".{now.microsecond // 1000:03d}"
            )
        else:
            time_str = now.strftime("%Y-%m-%d %H:%M:%S")

        method_upper = method.upper()
        if self.use_prefix_color:
            method_str = logstr.colored_str(method_upper, method)
        else:
            method_str = method_upper

        prefix_str = f"[{time_str}] [{method_str}] [{self.name}] "

        if self.use_prefix and self.use_prefix_color:
            prefix_str = logstr.colored_str(prefix_str, method)

        return prefix_str

    def log(
        self,
        method,
        msg,
        indent=0,
        fill=False,
        fill_side="both",
        end="\n",
        verbose: bool = None,
        *args,
        **kwargs,
    ):
        verbose = self.verbose if verbose is None else verbose
        if not verbose:
            return

        if type(msg) == str:
            msg_str = msg
        else:
            msg_str = repr(msg)
            quotes = ["'", '"']
            if msg_str[0] in quotes and msg_str[-1] in quotes:
                msg_str = msg_str[1:-1]

        if self.use_prefix:
            prefix_str = self.get_prefix_str(method)
        else:
            prefix_str = ""

        level, color = LOG_METHOD_COLORS[method]

        indent_str = " " * (self.log_indent + indent)
        indented_msg = "\n".join(
            [
                indent_str + prefix_str + logstr.colored_str(line, method)
                for line in msg_str.split("\n")
            ]
        )

        if fill:
            indented_msg = add_fills(indented_msg, fill_side=fill_side)

        handler = self.handlers[0]
        handler.terminator = end

        getattr(self, level)(indented_msg, *args, **kwargs)

    def route_log(self, method, msg, *args, **kwargs):
        level, color = LOG_METHOD_COLORS[method]
        # if level is lower (less important) than self.log_level, do not log
        if self.LEVEL_NAMES[level] < self.LEVEL_NAMES[self.log_level]:
            return
        functools.partial(self.log, method, msg)(*args, **kwargs)

    def err(self, msg: str = "", *args, **kwargs):
        self.route_log("err", msg, *args, **kwargs)

    def erro(self, msg: str = "", *args, **kwargs):
        self.route_log("erro", msg, *args, **kwargs)

    def warn(self, msg: str = "", *args, **kwargs):
        self.route_log("warn", msg, *args, **kwargs)

    def glow(self, msg: str = "", *args, **kwargs):
        self.route_log("glow", msg, *args, **kwargs)

    def hint(self, msg: str = "", *args, **kwargs):
        self.route_log("hint", msg, *args, **kwargs)

    def note(self, msg: str = "", *args, **kwargs):
        self.route_log("note", msg, *args, **kwargs)

    def mesg(self, msg: str = "", *args, **kwargs):
        self.route_log("mesg", msg, *args, **kwargs)

    def file(self, msg: str = "", *args, **kwargs):
        self.route_log("file", msg, *args, **kwargs)

    def line(self, msg: str = "", *args, **kwargs):
        self.route_log("line", msg, *args, **kwargs)

    def success(self, msg: str = "", *args, **kwargs):
        self.route_log("success", msg, *args, **kwargs)

    def okay(self, msg: str = "", *args, **kwargs):
        self.route_log("okay", msg, *args, **kwargs)

    def fail(self, msg: str = "", *args, **kwargs):
        self.route_log("fail", msg, *args, **kwargs)

    def back(self, msg: str = "", *args, **kwargs):
        self.route_log("back", msg, *args, **kwargs)

    def dbug(self, msg: str = "", *args, **kwargs):
        self.route_log("dbug", msg, *args, **kwargs)

    class TempIndent:
        def __init__(self, logger: "TCLogger", indent=2):
            self.logger: "TCLogger" = logger
            self.indent = indent

        def __enter__(self):
            self.logger.store_indent()
            self.logger.indent(self.indent)
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.logger.restore_indent()

    def temp_indent(self, indent=2):
        return self.TempIndent(self, indent=indent)


logger = TCLogger()


def log_error(mesg: str, error_type: type = ValueError):
    logger.warn(mesg)
    raise error_type(mesg)



import os
import re

from copy import deepcopy
from functools import partial
from pathlib import Path
from rapidfuzz import fuzz
from typing import Literal, Union, Protocol, Iterator

from .logs import TCLogger
from .types import KeysType, StrsType, PathType, PathsType
from .paths import norm_path

logger = TCLogger(__name__)


def match_val(
    val: str,
    vals: list[str],
    ignore_case: bool = True,
    spaces_to: Literal["keep", "ignore" "merge"] = "merge",
    use_fuzz: bool = False,
) -> tuple[str, int, float]:
    """
    Return:
        - closest val
        - index of the closest val in the list
        - similarity score (0-1)

    score = (1 – d/L)
    * d: distance by insert(1)/delete(1)/replace(2)
    * L: length sum of both strings
    """
    if not vals:
        return None, None, 0

    xval = deepcopy(val)
    xvals = deepcopy(vals)

    if spaces_to == "ignore":
        xval = re.sub(r"\s+", "", val.strip())
        xvals = [re.sub(r"\s+", "", v.strip()) for v in vals]
    elif spaces_to == "merge":
        xval = re.sub(r"\s+", " ", val.strip())
        xvals = [re.sub(r"\s+", " ", v.strip()) for v in vals]
    else:
        pass

    if ignore_case:
        xval = xval.lower()
        xvals = [v.lower() for v in xvals]

    if use_fuzz:
        scores = [fuzz.ratio(xval, v) / 100.0 for v in xvals]
    else:
        scores = [1 if xval == v else 0 for v in xvals]

    midx, max_score = None, 0.0
    for i, s in enumerate(scores):
        if s > max_score:
            midx = i
            max_score = s
    if midx is None:
        mval = None
    else:
        mval = vals[midx]
    return mval, midx, max_score


class MatchKeyFuncType(Protocol):
    """Match key with pattern.

    Example functions::
    - `match_key`: match key with pattern
    """

    def __call__(self, key: KeysType, pattern: KeysType, **kwargs) -> bool: ...


def unify_key_to_list(
    key: KeysType, ignore_case: bool = False, sep: str = "."
) -> list[Union[str, int]]:
    xkey = deepcopy(key)
    if isinstance(xkey, str):
        # split "a.b.c" to ["a", "b", "c"]
        xkey = xkey.split(sep)
    elif isinstance(xkey, (list, tuple)):
        # split case like ["a.b", "c"]
        xkey = [k.split(sep) if isinstance(k, str) else k for k in xkey]
        xkey = [item for sublist in xkey for item in sublist]
    if ignore_case:
        xkey = [k.lower() if isinstance(k, str) else k for k in xkey]
    return xkey


def unify_key_to_str(key: KeysType, ignore_case: bool = False, sep: str = ".") -> str:
    """Compared to `unitfy_key_to_list()`, this would enable match_key to accept str-format list idx in as key part.
    This means that "a.0.x" is same to ["a", 0, "x"] and ["a", "0", "x"].
    """
    xkey = deepcopy(key)
    if isinstance(xkey, (list, tuple)):
        xkey = sep.join(str(k) for k in xkey)
    if ignore_case:
        xkey = xkey.lower()
    return xkey


def match_key(
    key: KeysType,
    pattern: KeysType,
    ignore_case: bool = False,
    use_regex: bool = False,
    sep: str = ".",
) -> bool:
    unify_params = {"ignore_case": ignore_case, "sep": sep}
    xkey = unify_key_to_str(key, **unify_params)
    xpattern = unify_key_to_str(pattern, **unify_params)
    if use_regex:
        xpattern = rf"{xpattern}$"
        return re.search(xpattern, xkey) is not None
    else:
        return xkey == xpattern


def unify_paths(paths: PathsType) -> list[PathType]:
    """Unify paths to a list of Path objects."""
    if not paths:
        return []
    if isinstance(paths, (str, Path)):
        paths = [paths]
    return paths


def patternize_path(path: str) -> str:
    path = path.strip()
    # ignore comment lines
    if path.startswith("#"):
        return ""
    if path:
        # replace * to .*, and . (not ends with *) to \.
        path = re.sub(r"\*", ".*", path.strip())
        path = re.sub(r"\.(?!\*)", r"\.", path)
        # if not starts with /, prepend .*/
        if not path.startswith("/"):
            path = f".*/{path}"
        # remove trailing slashes
        path = path.rstrip("/")
    return path


def get_gitignore_patterns(root: PathType) -> list[str]:
    """Load .gitignore patterns.
    NOTE: Currently only covers .gitignore **under root**,
    and would support sub-directories in the future.
    """
    res = [".git"]
    path = Path(root) / ".gitignore"
    if path.exists():
        with path.open() as f:
            lines = f.read().splitlines()
        lines = [l.strip() for l in lines if l.strip() and not l.startswith("#")]
        res.extend(lines)
    return res


def unify_includes_excludes(
    root: PathType = ".",
    includes: StrsType = None,
    excludes: StrsType = None,
    use_gitignore: bool = True,
) -> tuple[list[str], list[str]]:
    """Unify includes and excludes to list of strs."""
    includes = unify_paths(includes)
    excludes = unify_paths(excludes)
    if use_gitignore:
        gitignore_patterns = get_gitignore_patterns(root)
        excludes.extend(gitignore_patterns)
    includes = [patternize_path(p) for p in includes]
    excludes = [patternize_path(p) for p in excludes]
    includes = [p for p in includes if p]
    excludes = [p for p in excludes if p]
    return includes, excludes


class InExCludeMatchFuncType(Protocol):
    """Match path with include or exclude pattern.

    Example functions::
    - `re_search`: match path with re.search
    """

    def __call__(self, path: PathType, pattern: str, **kwargs) -> bool: ...


def re_search(
    path: PathType,
    pattern: str,
    ignore_case: bool = True,
    suffix_match: bool = True,
) -> bool:
    """NOTE: Compared to `re.search(<pattern>, <string>, flags)`,
    the order of params `path` and `pattern` in `re_search` is swapped.
    The intension is to be consistent with `match_path()` and `match_key()`.
    """
    if ignore_case:
        flags = re.IGNORECASE
    else:
        flags = 0
    if suffix_match:
        pattern = rf"{pattern}$"
    return re.search(pattern, str(path), flags=flags) is not None


def inexclude_path_match(
    path: PathType,
    includes: StrsType = None,
    excludes: StrsType = None,
    include_match_func: InExCludeMatchFuncType = None,
    exclude_match_func: InExCludeMatchFuncType = None,
    unmatch_bool: bool = True,
) -> bool:
    if not includes and not excludes:
        return False

    if excludes:
        for exclude_pattern in excludes:
            if exclude_match_func(path, exclude_pattern):
                return False

    if includes:
        for include_pattern in includes:
            if include_match_func(path, include_pattern):
                return True

    # When program reaches here,
    # it means both includes and excludes are not matched

    # if includes is provided while excludes not, then unmatch means False
    if includes and not excludes:
        return False
    # if excludes is provided while includes not, then unmatch means True
    if excludes and not includes:
        return True
    # if both are provided, return unmatch_bool
    return unmatch_bool


class MatchPathFuncType(Protocol):
    """Match path with includes and excludes.

    Required params:
    - `path`: path to match
    - `includes`: list of include patterns
    - `excludes`: list of exclude patterns

    Example functions:
    - `match_path`: match path with includes and excludes
    """

    def __call__(
        self, path: PathType, includes: StrsType, excludes: StrsType, **kwargs
    ) -> bool: ...


def match_path(
    path: PathType,
    includes: StrsType = None,
    excludes: StrsType = None,
    unmatch_bool: bool = True,
    ignore_case: bool = True,
) -> bool:
    partial_params = {
        "ignore_case": ignore_case,
        "suffix_match": True,
    }
    include_match_func: InExCludeMatchFuncType = partial(re_search, **partial_params)
    exclude_match_func: InExCludeMatchFuncType = partial(re_search, **partial_params)
    return inexclude_path_match(
        path,
        includes=includes,
        excludes=excludes,
        include_match_func=include_match_func,
        exclude_match_func=exclude_match_func,
        unmatch_bool=unmatch_bool,
    )


def inner_iterate_folder(
    root: PathType = ".",
    match_func: MatchPathFuncType = match_path,
    verbose: bool = True,
    indent: int = 2,
    level: int = 0,
) -> Iterator[tuple[Path, bool, int]]:
    root = norm_path(root)
    if level == 0:
        logger.note(f"> {root}", verbose=verbose, indent=indent)
        temp_indent = indent + 2
    else:
        temp_indent = 2
    with logger.temp_indent(temp_indent):
        if root.is_file():
            yield root, match_func(root), level
        else:
            level = level + 1
            for p in os.listdir(root):
                p = Path(root) / p
                match_bool = match_func(p)
                if p.is_file():
                    yield p, match_bool, level
                elif p.is_dir():
                    if match_bool:
                        logger.note(f"> {p}", verbose=verbose)
                        yield p, match_bool, level
                        yield from inner_iterate_folder(
                            p,
                            match_func=match_func,
                            verbose=verbose,
                            level=level,
                        )
                    else:
                        yield p, match_bool, level
                else:
                    logger.warn(f"* skip: {p}", verbose=verbose)
                    pass


def iterate_folder(
    root: PathType = ".",
    includes: StrsType = None,
    excludes: StrsType = None,
    yield_folder: bool = False,
    unmatch_bool: bool = True,
    ignore_case: bool = True,
    use_gitignore: bool = True,
    verbose: bool = True,
    indent: int = 2,
) -> Iterator[tuple[Path, bool, int]]:
    """Iterate paths with includes and excludes.

    Args:
        - root: root path to traverse
        - includes: list of patterns to include
        - excludes: list of patterns to exclude
        - unmatch_bool: default bool when path is not matched by neither includes nor excludes
        - ignore_case: ignore case when matching
        - use_gitignore: whether to add .gitignore to excludes
        - verbose: whether to log process
        - indent: logging indent spaces

    Yields:
        - (path, match_bool, level)

    In a previous version, there are two more args:
        - include_func: function to call when a path is included
        - exclude_func: function to call when a path is excluded
    These two args are removed, as this target could be reached more easily, by just iterate the iterator and operate on the yielded (path, match_bool, level) tuples.
    """
    includes, excludes = unify_includes_excludes(
        root, includes=includes, excludes=excludes, use_gitignore=use_gitignore
    )
    match_func: MatchPathFuncType = partial(
        match_path,
        includes=includes,
        excludes=excludes,
        unmatch_bool=unmatch_bool,
        ignore_case=ignore_case,
    )
    for p, match_bool, level in inner_iterate_folder(
        root, match_func=match_func, verbose=verbose, indent=indent
    ):
        if not yield_folder and p.is_dir():
            continue
        if verbose and p.is_file():
            if match_bool:
                logger.file(f"+ {p}")
            if not match_bool:
                logger.warn(f"- {p}")
        yield p, match_bool, level


def match_paths(
    root: PathType = ".",
    includes: StrsType = None,
    excludes: StrsType = None,
    unmatch_bool: bool = True,
    ignore_case: bool = True,
    use_gitignore: bool = True,
    to_str: bool = True,
    verbose: bool = True,
    indent: int = 2,
) -> dict:
    """Match paths in a folder with includes and excludes.
    Args:
        - to_str: whether to convert paths to str in result
        - See `iterate_folder()` for other arguments.

    Return dict with keys:

    ```py
    {
        "includes": <list of matched include paths>,
        "excludes": <list of matched exclude paths>,
    }
    ```
    """

    res = {
        "includes": [],
        "excludes": [],
    }
    for p, match_bool, level in iterate_folder(
        root,
        includes=includes,
        excludes=excludes,
        unmatch_bool=unmatch_bool,
        ignore_case=ignore_case,
        use_gitignore=use_gitignore,
        verbose=verbose,
        indent=indent,
    ):
        if match_bool:
            res["includes"].append(p)
        else:
            res["excludes"].append(p)

    if to_str:
        res["includes"] = [str(p) for p in res["includes"]]
        res["excludes"] = [str(p) for p in res["excludes"]]

    return res



from collections.abc import Callable

"""It is ok to use lambdas to shorten codes of calls,
but readability and extensibility are more important here.
"""

# helper functions


def type_str(attr) -> str:
    attr_type = type(attr)
    type_module = attr_type.__module__
    type_name = attr_type.__name__
    if type_module == "builtins":
        return type_name
    else:
        return f"{type_module}.{type_name}"


def dict_type_str(obj: dict) -> str:
    dict_keys = str(list(obj.keys()))
    dict_keys_str = f"{{{dict_keys[1:-1]}}}"
    return f"dict{dict_keys_str}[{len(obj)}]"


# default is_include in attrs_to_dict


def is_attr_included(name: str, attr) -> bool:
    """Only include attrs that are: non-internal, non-callable, and non-staticmethod"""
    return (
        not name.startswith("__")
        and not callable(attr)
        and type(attr) is not staticmethod
    )


# default callables in attrs_to_dict


def none_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return attr is None

    def val_call(name, attr):
        return "None"

    return key_call, val_call


def constant_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return isinstance(attr, (int, float, str, bool, type(None)))

    def val_call(name, attr):
        if isinstance(attr, str):
            return f"'{attr}'"
        if isinstance(attr, (int, float, bool)):
            return attr
        return f"({type_str(attr)}) {attr}"

    return key_call, val_call


def list_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return isinstance(attr, list)

    def val_call(name, attr):
        if len(attr) > 0:
            list_type = type(attr[0]).__name__
            list_size = len(attr)
            if list_type == "dict":
                return dict_type_str(attr[0])
            else:
                return f"list({list_type})[{list_size}]"
        else:
            return f"[]"

    return key_call, val_call


def tuple_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return isinstance(attr, tuple)

    def val_call(name, attr):
        if len(attr) > 0:
            tuple_types = str(list(map(lambda x: type(x).__name__, attr)))
            tuple_size = len(attr)
            tuple_types_str = f"({tuple_types[1:-1]})"
            return f"tuple{tuple_types_str}[{tuple_size}]"
        else:
            return f"(,)"

    return key_call, val_call


def dict_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return isinstance(attr, dict)

    def val_call(name, attr):
        return dict_type_str(attr)

    return key_call, val_call


def iter_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return hasattr(attr, "__iter__")

    def val_call(name, attr):
        if hasattr(attr, "__len__"):
            return f"iter[{len(attr)}]"
        else:
            return f"iter"

    return key_call, val_call


def type_call() -> tuple[Callable, Callable]:
    def key_call(name, attr) -> bool:
        return True

    def val_call(name, attr):
        return f"<{type_str(attr)}>"

    return key_call, val_call


MID_CALLS = [
    *[none_call(), constant_call()],
    *[list_call(), tuple_call(), dict_call(), iter_call()],
]
LAST_CALLS = [type_call()]


# main function
def attrs_to_dict(
    obj,
    is_include: Callable = is_attr_included,
    pre_calls: list[tuple[Callable, Callable]] = None,
    post_calls: list[tuple[Callable, Callable]] = None,
) -> dict:
    """The is_include, and each callable in pre_calls and post_calls should take two arguments:
    - name: str, the name of the attribute
    - attr, the attribute itself
    The pre_calls are called before mid_calls, and the post_calls are called after mid_calls.

    For is_include:
    - if True, the attr would be included in the result
    - if False, the attribute would be ignored

    For each callable in pre_calls and post_calls:
    - the 1st callable (key_call) is to determine whether the 2nd callable (val_call) should be called
    - if key_call returns True, val_call would be called, and the result would be the value of the attr
    - if all key_call return False, the last_call would be called to set the default value of the attr
    """
    res = {}
    pre_calls = pre_calls or []
    post_calls = post_calls or []
    calls = [*pre_calls, *MID_CALLS, *post_calls, *LAST_CALLS]
    for name, attr in obj.__dict__.items():
        if is_include(name, attr):
            for key_call, val_call in calls:
                if key_call(name, attr):
                    res[name] = val_call(name, attr)
                    break

    return res



"""Format utils"""

from copy import deepcopy
from typing import Union, Literal

from .colors import colored, decolored
from .maths import max_key_len
from .strings import chars_slice, chars_len


class DictListAligner:
    def __init__(self):
        self.list_item_types = {}

    def get_item_type(self, v: list) -> Union[str, bool, Union[int, float]]:
        if not v:
            return None
        if isinstance(v[0], bool):
            return "bool"
        elif isinstance(v[0], (int, float)):
            return "num"
        elif isinstance(v[0], str):
            return "str"
        else:
            return str(type(v[0]))

    def is_all_items_types_valid(self, lst: list) -> bool:
        for v in lst:
            if not isinstance(v, (bool, int, float, str)):
                return False
        return True

    def extract_same_len_lists(self, d: dict) -> dict[int, dict]:
        dict_lists_by_len: dict[int, dict] = {}
        for k, v in d.items():
            if isinstance(v, list):
                if not self.is_all_items_types_valid(v):
                    continue
                v_len = len(v)
                item_type = self.get_item_type(v)
                ak = (v_len, item_type)
                self.list_item_types[k] = item_type
                if ak not in dict_lists_by_len:
                    dict_lists_by_len[ak] = {}
                dict_lists_by_len[ak][k] = v
        return dict_lists_by_len

    def calc_aligned_widths(
        self, d: dict
    ) -> dict[tuple[int, Union[str, bool, Union[int, float]]], list]:
        aligned_widths_by_len: dict[tuple, list] = {}
        for k, v in d.items():
            if isinstance(v, list):
                if not self.is_all_items_types_valid(v):
                    continue
                list_len = len(v)
                item_type = self.get_item_type(v)
                ak = (list_len, item_type)
                if ak not in aligned_widths_by_len:
                    aligned_widths_by_len[ak] = [0] * list_len
                for i, vv in enumerate(v):
                    if isinstance(vv, str):
                        vv_width = chars_len(decolored(vv))
                    else:
                        vv_width = chars_len(str(vv))
                    aligned_widths_by_len[ak][i] = max(
                        aligned_widths_by_len[ak][i], vv_width
                    )
        return aligned_widths_by_len

    def align_lists_in_dict(self, d: dict, align: Literal["l", "r"] = None) -> None:
        dict_lists_by_len = self.extract_same_len_lists(d)
        aligned_widths_by_len = self.calc_aligned_widths(d)
        for ak, dd in dict_lists_by_len.items():
            list_len, item_type = ak
            aligned_widths = aligned_widths_by_len[ak]
            for k, v in dd.items():
                for i, vv in enumerate(v):
                    v[i] = chars_slice(vv, end=aligned_widths[i], align=align)
                d[k] = v


class DictStringifier:
    def __init__(
        self,
        indent: int = 2,
        max_depth: int = None,
        align_colon: bool = True,
        align_list: bool = True,
        align_list_side: Literal["l", "r"] = None,
        add_quotes: bool = True,
        is_colored: bool = True,
        brace_colors: list[str] = ["light_blue", "light_cyan", "light_magenta"],
        key_colors: list[str] = ["light_blue", "light_cyan", "light_magenta"],
        value_colors: list[str] = ["white"],
    ):
        self.indent = indent
        self.max_depth = max_depth
        self.align_colon = align_colon
        self.align_list = align_list
        self.align_list_side = align_list_side
        self.add_quotes = add_quotes
        self.is_colored = is_colored
        self.depth_configs = {}
        self.brace_colors = brace_colors
        self.key_colors = key_colors
        self.value_colors = value_colors

    def get_depth_config(self, depth: int):
        if depth in self.depth_configs:
            return self.depth_configs[depth]

        if isinstance(self.key_colors, str):
            key_color = self.key_colors
        else:
            key_color = self.key_colors[depth % len(self.key_colors)]
        if isinstance(self.value_colors, str):
            value_color = self.value_colors
        else:
            value_color = self.value_colors[depth % len(self.value_colors)]
        if isinstance(self.brace_colors, str):
            brace_color = self.brace_colors
        else:
            brace_color = self.brace_colors[depth % len(self.brace_colors)]

        indent_str = " " * self.indent * (depth + 1)
        brace_indent_str = " " * self.indent * depth
        if self.is_colored:
            lb = colored("{", brace_color)
            rb = colored("}", brace_color)
            lk = colored("[", brace_color)
            rk = colored("]", brace_color)
            colon = colored(":", brace_color)
            comma = colored(",", brace_color)
            ellipsis = colored("...", value_color)
        else:
            lb, rb = "{", "}"
            lk, rk = "[", "]"
            colon = ":"
            comma = ","
            ellipsis = "..."

        self.depth_configs[depth] = {
            "key_color": key_color,
            "value_color": value_color,
            "brace_color": brace_color,
            "indent_str": indent_str,
            "brace_indent_str": brace_indent_str,
            "lb": lb,
            "rb": rb,
            "lk": lk,
            "rk": rk,
            "colon": colon,
            "comma": comma,
            "ellipsis": ellipsis,
        }

        return self.depth_configs[depth]

    def dict_to_str(
        self,
        d: Union[dict, list],
        depth: int = 0,
    ) -> tuple[str, str]:
        configs = self.get_depth_config(depth)
        key_color = configs["key_color"]
        value_color = configs["value_color"]
        indent_str = configs["indent_str"]
        brace_indent_str = configs["brace_indent_str"]
        lb, rb = configs["lb"], configs["rb"]
        lk, rk = configs["lk"], configs["rk"]
        colon = configs["colon"]
        comma = configs["comma"]
        ellipsis = configs["ellipsis"]

        if self.max_depth is not None and depth > self.max_depth:
            return f"{lb}{ellipsis}{rb}", "dict"

        lines = []
        if isinstance(d, dict):
            if self.align_list:
                aligner = DictListAligner()
                aligner.align_lists_in_dict(d, align=self.align_list_side)
            if self.add_quotes:
                key_len = max_key_len(d, 2)
            else:
                key_len = max_key_len(d)
            for idx, (key, value) in enumerate(d.items()):
                key_str = f"{key}"
                if self.add_quotes:
                    key_str = f'"{key_str}"'
                if self.align_colon:
                    key_str = chars_slice(key_str, end=key_len)
                value_str, value_str_type = self.dict_to_str(
                    value, depth=depth + 1 if isinstance(value, (dict, list)) else depth
                )
                if self.add_quotes:
                    if isinstance(value_str, str) and value_str_type == "str":
                        value_str = f'"{value_str}"'
                if self.is_colored:
                    colored_key_str = colored(key_str, key_color)
                    colored_value_str = colored(value_str, value_color)
                    line = f"{indent_str}{colored_key_str} {colon} {colored_value_str}"
                else:
                    line = f"{indent_str}{key_str} {colon} {value_str}"
                if self.align_list:
                    if key in aligner.list_item_types:
                        if aligner.list_item_types[key] != "str":
                            line = line.replace("'", "")
                if idx < len(d) - 1:
                    line += comma
                lines.append(line)
            if lines:
                lines_str = "\n".join(lines)
                dict_str = f"{lb}\n{lines_str}\n{brace_indent_str}{rb}"
            else:
                dict_str = f"{lb}{rb}"
            str_type = "dict"
        elif isinstance(d, list):
            is_list_contain_dict = any(isinstance(v, dict) for v in d)
            if is_list_contain_dict:
                list_strs = []
                for v in d:
                    v_str = self.dict_to_str(v, depth=depth)[0]
                    list_strs.append(v_str)
                dict_str = f"{lk}{', '.join(list_strs)}{rk}"
            else:
                dict_str = [self.dict_to_str(v, depth=depth)[0] for v in d]
            str_type = "list"
        else:
            dict_str = d
            str_type = "str"

        return dict_str, str_type


def dict_to_str(
    d: dict,
    indent: int = 2,
    max_depth: int = None,
    align_colon: bool = True,
    align_list: bool = True,
    align_list_side: Literal["l", "r"] = None,
    add_quotes: bool = False,
    is_colored: bool = True,
    brace_colors: list[str] = ["light_blue", "light_cyan", "light_magenta"],
    key_colors: list[str] = ["light_blue", "light_cyan", "light_magenta"],
    value_colors: list[str] = ["white"],
) -> str:
    ds = DictStringifier(
        indent=indent,
        max_depth=max_depth,
        align_colon=align_colon,
        align_list=align_list,
        align_list_side=align_list_side,
        add_quotes=add_quotes,
        is_colored=is_colored,
        brace_colors=brace_colors,
        key_colors=key_colors,
        value_colors=value_colors,
    )
    return ds.dict_to_str(deepcopy(d))[0]



from pathlib import Path
from typing import Literal

from .types import PathType, PathsType, StrsType
from .paths import norm_path, strf_path
from .matches import iterate_folder
from .logs import logstr

CON_PIPE = "─"  # pipe of connect line
MID_PIPE = "├"  # pipe of files not the last
END_PIPE = "└"  # pipe of files the last
IND_PIPE = "│"  # pipe of indent before each file
FOLD_BULLET = ">"  # bullet for folder
FILE_BULLET = "*"  # bullet for file


class PathObj:
    def __init__(self, path: Path, match_bool: bool, level: int):
        self.path = path
        self.match_bool = match_bool
        self.level = level
        self.idx = None
        self.prev: "PathObj" = None
        self.next: "PathObj" = None
        self.parent = None
        self.children: list["PathObj"] = []
        self.is_children_sorted = False
        self.is_last_obj = None  # last folder or last file
        self.files_count = None  # only for folder
        self.lines_count = None  # only for file

    def is_file(self) -> bool:
        return self.path.is_file()

    def is_folder(self) -> bool:
        return self.path.is_dir()

    def full_str(self) -> str:
        return strf_path(self.path)

    def full_str_with_slash(self) -> str:
        path_str = self.full_str()
        if self.is_folder():
            path_str += "/"
        return path_str

    def short_str(self) -> str:
        return self.path.name

    def short_str_with_slash(self) -> str:
        path_str = self.short_str()
        if self.is_folder():
            path_str += "/"
        return path_str

    def set_idx(self, idx: int):
        self.idx = idx

    def set_prev(self, prev: "PathObj"):
        self.prev = prev

    def set_next(self, next: "PathObj"):
        self.next = next

    def set_parent(self, parent: "PathObj"):
        self.parent = parent

    def append_child(self, child: "PathObj"):
        self.children.append(child)
        self.is_children_sorted = False

    def get_last_child(self) -> "PathObj":
        if not self.children:
            return None
        if not self.is_children_sorted:
            self.children.sort(key=lambda c: c.full_str_with_slash())
            self.is_children_sorted = True
        return self.children[-1]

    def set_is_last_obj(self):
        self.is_last_obj = True

    def get_closest_prev_folder(self) -> "PathObj":
        if not self.prev:
            return None
        if self.prev.is_folder():
            return self.prev
        else:
            return self.prev.get_closest_prev_folder()

    def get_closest_next_folder(self) -> "PathObj":
        if not self.next:
            return None
        if self.next.is_folder():
            return self.next
        else:
            return self.next.get_closest_next_folder()

    def get_closest_prev_file(self) -> "PathObj":
        if not self.prev:
            return None
        if self.prev.is_file():
            return self.prev
        else:
            return self.prev.get_closest_prev_file()

    def get_closest_next_file(self) -> "PathObj":
        if not self.next:
            return None
        if self.next.is_file():
            return self.next
        else:
            return self.next.get_closest_next_file()


def get_path_obj_parent(
    path_obj: PathObj,
    path_objs: list[PathObj],
    beg_idx: int = None,
    end_idx: int = None,
) -> PathObj:
    if beg_idx is None:
        beg_idx = 0
    if end_idx is None:
        end_idx = len(path_objs)
    for i in range(end_idx - 1, beg_idx - 1, -1):
        po = path_objs[i]
        if po.path == path_obj.path.parent:
            return po
    return None


def path_obj_to_str(
    path_obj: PathObj,
    name_style: Literal["short", "full"] = "short",
    prefix_style: Literal["pipe", "bullet", "space"] = "pipe",
    branch_style: Literal["pipe", "bullet", "space"] = "pipe",
    show_file_lines_count: bool = True,
    show_folder_files_count: bool = True,
    show_folder_slash: bool = True,
    show_color: bool = True,
    indent_width: int = 4,
) -> str:
    level = path_obj.level

    # path_str
    if name_style == "short":
        path_str = path_obj.short_str()
    else:
        path_str = path_obj.full_str()
    if show_folder_slash and path_obj.is_folder():
        path_str += "/"

    # init before_strs
    before_strs = [" "] * (level * indent_width)

    # set branch chars (| before prefix)
    if level >= 2 and branch_style == "pipe":
        branch_bools = []
        parent = path_obj.parent
        while parent:
            if parent.is_last_obj:
                branch_bools.insert(0, False)
            else:
                branch_bools.insert(0, True)
            parent = parent.parent
        for idx, branch_bool in enumerate(branch_bools):
            if branch_bool:
                before_strs[(idx - 1) * indent_width] = IND_PIPE
    else:
        pass

    # set prefix chars (chars before path_str)
    if level >= 1 and prefix_style == "pipe":
        if path_obj.is_last_obj:
            before_strs[-indent_width] = END_PIPE
        else:
            before_strs[-indent_width] = MID_PIPE
        if indent_width > 2:
            before_strs[-indent_width + 1 : -1] = [CON_PIPE] * (indent_width - 2)
    elif prefix_style == "bullet":
        if path_obj.is_folder():
            before_strs[-indent_width] = FOLD_BULLET
        else:  # path_obj.is_file()
            before_strs[-indent_width] = FILE_BULLET
    else:
        pass

    # join before_strs
    before_str = "".join(before_strs)

    # set path_obj_str
    path_obj_str = f"{before_str}{path_str}"

    # set color
    if show_color:
        if path_obj.match_bool is False:
            path_obj_str = logstr.warn(path_obj_str)
        elif path_obj.is_folder():
            path_obj_str = logstr.note(path_obj_str)
        elif path_obj.is_file():
            path_obj_str = logstr.file(path_obj_str)

    return path_obj_str


def tree_folder(
    # params of paths matching
    root: PathType = ".",
    includes: StrsType = None,
    excludes: StrsType = None,
    unmatch_bool: bool = True,
    ignore_case: bool = True,
    use_gitignore: bool = True,
    # params of tree formatting
    name_style: Literal["short", "full"] = "short",
    prefix_style: Literal["pipe", "bullet", "space"] = "pipe",
    branch_style: Literal["pipe", "bullet", "space"] = "pipe",
    show_file_lines_count: bool = True,
    show_folder_files_count: bool = True,
    show_folder_slash: bool = True,
    show_color: bool = True,
    indent_width: int = 4,
    max_level: int = None,
    verbose: bool = True,
) -> str:
    """
    Example output:

    tclogger/
    ├── example.py
    ├── pyproject.toml
    ├── README.md
    ├── src/
    │   ├── tclogger/
    │   │   ├── trees.py
    |   │   ├── ...
    │   │   └── types.py
    │   └── tclogger.egg-info/
    │       ├── ...
    │       └── top_level.txt
    └── test.log
    """
    # create path objs
    root = norm_path(root)
    path_objs: list[PathObj] = []
    root_obj = PathObj(root, True, 0)
    path_objs.append(root_obj)
    p: PathType
    for p, match_bool, level in iterate_folder(
        root,
        includes=includes,
        excludes=excludes,
        yield_folder=True,
        unmatch_bool=unmatch_bool,
        ignore_case=ignore_case,
        use_gitignore=use_gitignore,
        verbose=False,
        indent=0,
    ):
        path_obj = PathObj(p, match_bool, level)
        path_objs.append(path_obj)

    # sort path objs by full_str
    path_objs = sorted(path_objs, key=lambda po: po.full_str_with_slash())

    # set idx, prev, next
    prev: PathObj = None
    for idx, path_obj in enumerate(path_objs):
        path_obj.set_idx(idx)
        path_obj.set_prev(prev)
        if prev:
            prev.set_next(path_obj)
        prev = path_obj

    # set parent
    for idx, path_obj in enumerate(path_objs):
        parent = get_path_obj_parent(path_obj, path_objs, end_idx=idx)
        path_obj.set_parent(parent)
        if parent:
            parent.append_child(path_obj)

    # set is_last_obj
    for idx, path_obj in enumerate(path_objs):
        if path_obj.is_folder():
            last_child = path_obj.get_last_child()
            if last_child:
                last_child.set_is_last_obj()

    path_obj_to_str_params = {
        "name_style": name_style,
        "prefix_style": prefix_style,
        "branch_style": branch_style,
        "show_file_lines_count": show_file_lines_count,
        "show_folder_files_count": show_folder_files_count,
        "show_folder_slash": show_folder_slash,
        "show_color": show_color,
        "indent_width": indent_width,
    }

    # format path objs to strings
    path_obj_strs = []
    for path_obj in path_objs:
        path_obj_str = path_obj_to_str(path_obj, **path_obj_to_str_params)
        path_obj_strs.append(path_obj_str)

    if verbose:
        for path_obj_str in path_obj_strs:
            print(path_obj_str)



from pathlib import Path

from .types import PathType


def norm_path(path: PathType) -> Path:
    """Normalize path/str (expanduser +resolve). Return Path type."""
    if isinstance(path, str):
        path = Path(path)
    return path.expanduser().resolve()


def strf_path(path: PathType) -> str:
    """Normalize path/str (expanduser +resolve). Return str type."""
    return str(norm_path(path))



"""Time utils"""

from datetime import datetime, timedelta
from typing import Literal, Union
from zoneinfo import ZoneInfo

TIMEZONE = "Asia/Shanghai"


def set_timezone(tz: str = "Asia/Shanghai") -> None:
    global TIMEZONE
    TIMEZONE = tz


class tcdatetime(datetime):
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        return instance.replace(tzinfo=ZoneInfo(TIMEZONE))

    @classmethod
    def now(cls):
        return datetime.now(ZoneInfo(TIMEZONE))

    @classmethod
    def fromtimestamp(ts):
        return datetime.fromtimestamp(ts, ZoneInfo(TIMEZONE))

    @classmethod
    def fromisoformat(cls, s):
        return datetime.fromisoformat(s).replace(tzinfo=ZoneInfo(TIMEZONE))

    @classmethod
    def min(cls):
        """NOTE: In datetime, it should be `datetime.min`. However, classmethod property is no longer supported for Python >= 3.13, so use `tcdatetime.min()`."""
        return datetime.min.replace(tzinfo=ZoneInfo(TIMEZONE))

    @classmethod
    def max(cls):
        """NOTE: In datetime, it should be `datetime.max`. However, classmethod property is no longer supported for Python >= 3.13, so use `tcdatetime.max()`."""
        return datetime.max.replace(tzinfo=ZoneInfo(TIMEZONE))


def get_now() -> datetime:
    return datetime.now(ZoneInfo(TIMEZONE))


def get_now_ts() -> int:
    return int(get_now().timestamp())


def get_now_str() -> str:
    return get_now().strftime("%Y-%m-%d %H:%M:%S")


def get_now_ts_str() -> tuple[int, str]:
    now = get_now()
    now_ts = int(now.timestamp())
    now_str = now.strftime("%Y-%m-%d %H:%M:%S")
    return now_ts, now_str


def get_date_str(date_str: str = None) -> str:
    return date_str or get_now_str()[:10]


def ts_to_str(ts: int) -> str:
    return datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")


def str_to_t(s: str) -> datetime:
    return datetime.fromisoformat(s)


def str_to_ts(s: str) -> int:
    return int(datetime.fromisoformat(s).timestamp())


def t_to_str(t: datetime) -> str:
    return t.strftime("%Y-%m-%d %H:%M:%S")


def t_to_ts(t: datetime) -> int:
    return int(t.timestamp())


def dt_to_sec(dt: timedelta, precision: int = 0) -> float:
    if precision is not None and precision > 0:
        return round(dt.total_seconds(), ndigits=precision)
    else:
        return int(dt.total_seconds())


def dt_to_str(
    dt: Union[timedelta, int, float],
    precision: int = 0,
    str_format: Literal["unit", "colon"] = "colon",
) -> str:
    if isinstance(dt, timedelta):
        hours = dt.days * 24 + dt.seconds // 3600
        minutes = (dt.seconds // 60) % 60
        seconds = dt.seconds % 60
        microseconds = dt.microseconds / 1000000
        precised_seconds = seconds + microseconds
    else:
        hours = int(dt) // 3600
        minutes = (int(dt) // 60) % 60
        seconds = int(dt) % 60
        microseconds = dt - int(dt)
        precised_seconds = seconds + microseconds

    if str_format == "unit":
        hours_str = f"{hours}hr" if hours > 0 else ""
        minutes_str = f"{minutes}min" if minutes > 0 else ""
        if precision is not None and precision > 0:
            seconds_str = f"{precised_seconds:.{precision}f}s"
        else:
            seconds_str = f"{seconds}s"

        time_str = " ".join([hours_str, minutes_str, seconds_str]).strip()
    else:
        hours_str = f"{hours:02d}" if hours > 0 else ""
        minutes_str = f"{minutes:02d}"
        seconds_str = f"{seconds:02d}"
        time_str = ":".join([hours_str, minutes_str, seconds_str]).strip(":")
        if precision is not None and precision > 0:
            time_str += f".{int(microseconds * 10**precision):0{precision}d}"

    return time_str


def unify_ts_and_str(
    t: Union[str, int, None]
) -> tuple[Union[int, None], Union[str, None]]:
    if t is None:
        return None, None
    if isinstance(t, str):
        return str_to_ts(t), t
    if isinstance(t, int):
        return t, ts_to_str(t)
    return t, t



import sys
import linecache

from .decorations import brk
from .logs import logstr
from .fills import add_fills


class BreakpointException(Exception):
    def __init__(self, *args, prev_n: int = 1, next_n: int = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.__suppress_context__ = True
        # Get the frame where raise_breakpoint() was called
        frame = sys._getframe(2)
        self.filename = frame.f_code.co_filename
        self.lineno = frame.f_lineno
        self.name = frame.f_code.co_name
        self.prev_n = prev_n
        self.next_n = next_n
        self.msg = args[0] if args else ""
        # Get class name if the function is a method
        self.class_name = None
        if "self" in frame.f_locals:
            self.class_name = frame.f_locals["self"].__class__.__name__
        elif "cls" in frame.f_locals:
            self.class_name = frame.f_locals["cls"].__name__

    def __str__(self):
        lines = []
        fills_str = logstr.warn(add_fills(filler="-"))
        lines.append(fills_str)

        if self.class_name:
            func_name = f"{self.class_name}.{self.name}():"
        else:
            func_name = f"{self.name}:"
        file_info_str = f"* File {logstr.file(brk(self.filename))}, line {logstr.file(self.lineno)}, in {logstr.mesg(func_name)}"
        lines.append(logstr.warn(file_info_str))

        beg_line_no = max(1, self.lineno - self.prev_n)
        end_line_no = self.lineno + self.next_n
        line_num_width = len(str(end_line_no))
        for line_num in range(beg_line_no, end_line_no + 1):
            line_content = linecache.getline(self.filename, line_num).rstrip()
            line_num_content = f"| {line_num:>{line_num_width}} |  {line_content}"
            if line_num == self.lineno:
                # Highlight breakpoint line
                lines.append(logstr.warn(line_num_content))
            else:
                lines.append(logstr.line(line_num_content))

        if self.msg:
            exp_str = f"× BreakpointException: {self.msg}"
        else:
            exp_str = f"× BreakpointException"
        lines.append(logstr.warn(exp_str))
        lines.append(fills_str)

        return "\n".join(lines)


def raise_breakpoint(msg: str = "", prev_n: int = 1, next_n: int = 1):
    exc = BreakpointException(msg, prev_n=prev_n, next_n=next_n)
    sys.excepthook = lambda exc_type, exc_value, exc_tb: print(
        str(exc_value), file=sys.stderr
    )
    raise exc



from pathlib import Path
from typing import Union

KeyType = Union[str, int]
KeysType = Union[KeyType, list[KeyType]]

PathType = Union[str, Path]
PathsType = Union[PathType, list[PathType]]

StrsType = Union[str, list[str]]
IntsType = Union[int, list[int]]

DictListType = Union[dict, list]

LIST_TYPES = (list, tuple, set)



from copy import deepcopy
from functools import partial

from .types import KeysType, DictListType
from .matches import MatchKeyFuncType, unify_key_to_list, match_key


class CaseInsensitiveDict(dict):
    """Inspired by: https://stackoverflow.com/a/32888599"""

    @classmethod
    def _k(cls, key):
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveDict, self).__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(self.__class__._k(key))

    def __setitem__(self, key, value):
        super(CaseInsensitiveDict, self).__setitem__(self.__class__._k(key), value)

    def __delitem__(self, key):
        return super(CaseInsensitiveDict, self).__delitem__(self.__class__._k(key))

    def __contains__(self, key):
        return super(CaseInsensitiveDict, self).__contains__(self.__class__._k(key))

    def has_key(self, key):
        return super(CaseInsensitiveDict, self).has_key(self.__class__._k(key))

    def pop(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).pop(
            self.__class__._k(key), *args, **kwargs
        )

    def get(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).get(
            self.__class__._k(key), *args, **kwargs
        )

    def setdefault(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).setdefault(
            self.__class__._k(key), *args, **kwargs
        )

    def update(self, E={}, **F):
        super(CaseInsensitiveDict, self).update(self.__class__(E))
        super(CaseInsensitiveDict, self).update(self.__class__(**F))

    def _convert_keys(self):
        for k in list(self.keys()):
            v = super(CaseInsensitiveDict, self).pop(k)
            self.__setitem__(k, v)


def dict_get(d: dict, keys: KeysType, default=None, sep: str = "."):
    if isinstance(keys, str) and sep:
        keys = keys.split(sep)
    for key in keys:
        if (isinstance(d, dict) and key in d) or (isinstance(d, list) and key < len(d)):
            d = d[key]
        else:
            return default
    return d


def dict_set(d: dict, keys: KeysType, value, sep: str = "."):
    if isinstance(keys, str) and sep:
        keys = keys.split(sep)
    for key in keys[:-1]:
        if isinstance(d, dict):
            d = d.setdefault(key, {})
        elif isinstance(d, list):
            if key >= len(d):
                d.extend([{} for _ in range(key - len(d) + 1)])
            d = d[key]

    if isinstance(d, dict):
        d[keys[-1]] = value
    elif isinstance(d, list):
        if keys[-1] >= len(d):
            d.extend([None for _ in range(keys[-1] - len(d) + 1)])
        d[keys[-1]] = value


def dict_get_all(
    d: DictListType,
    key: KeysType,
    value,
    ignore_case: bool = False,
    use_regex: bool = False,
    sep: str = ".",
    index_list: bool = False,
): ...


def inner_dict_set_all(
    d: DictListType,
    key: KeysType,
    value,
    index_list: bool,
    keys_path: list,
    match_func: MatchKeyFuncType = match_key,
):
    """recursive helper function for `dict_set_all()`"""
    if isinstance(d, dict):
        for k, v in d.items():
            new_keys_path = keys_path + [k]
            if match_func(new_keys_path, key):
                d[k] = value
            if isinstance(v, (dict, list)):
                inner_dict_set_all(
                    v,
                    key,
                    value,
                    index_list=index_list,
                    keys_path=new_keys_path,
                    match_func=match_func,
                )

    elif isinstance(d, list):
        for idx, item in enumerate(d):
            if index_list:
                new_keys_path = keys_path + [idx]
            else:
                new_keys_path = keys_path

            if match_func(new_keys_path, key):
                d[idx] = value

            if isinstance(item, (dict, list)):
                inner_dict_set_all(
                    item,
                    key,
                    value,
                    index_list=index_list,
                    keys_path=new_keys_path,
                    match_func=match_func,
                )


def dict_set_all(
    d: DictListType,
    key: KeysType,
    value,
    index_list: bool = False,
    ignore_case: bool = False,
    use_regex: bool = False,
    sep: str = ".",
):
    """
    Recursively replace values of all items with field `key` in a dict/list.
    Args:
    - d: input dict or list
    - key: key (pattern) to match and set
    - value: value to set for all matching keys
    - ignore_case: ignore case of key when matching
    - use_regex: use regex to match key
    - sep: separator to use for chaining keys
    - index_list: if True, include list indices in the keys path when matching
    """
    match_func = partial(
        match_key,
        ignore_case=ignore_case,
        use_regex=use_regex,
        sep=sep,
    )
    inner_dict_set_all(
        d,
        key,
        value,
        index_list=index_list,
        keys_path=[],
        match_func=match_func,
    )


def dict_pop(d: DictListType, key: str):
    res = None
    if isinstance(d, dict):
        if key in d:
            res = d.pop(key)
    elif isinstance(d, list):
        res = [dict_pop(item, key) for item in d]
    else:
        return None
    return res


def dict_extract(d: DictListType, keys: list, key_level: int = 0, pop: bool = False):
    """
    Inputs:
    - d: (dict or list)
    - keys: (list)
    - pop: (bool) remove the last key from d

    Example 1:
        - d: {"pages": [{"part": "part1"}, {"part": "part2"}]}
        - keys: ["pages", "part"]
        - output: ["part1", "part2"]
        - (if pop:
            d: {"pages": []}
        )
    Example 2:
        - d: {"owner": {"name": "name1", "mid": 123}}
        - keys: ["owner", "name"]
        - output: "name1"
        - (if pop:
            d: {"owner": {"mid": 123}}
        )
    """
    if not d or not keys[key_level:]:
        return d
    if isinstance(d, dict):
        level_key = keys[key_level]
        if level_key not in d:
            return None
        if key_level == 0 and len(keys) == 1:
            res = d.get(keys[0], None)
            if pop:
                dict_pop(d, keys[0])
            return res
        res = dict_extract(d[level_key], keys, key_level=key_level + 1, pop=pop)
        if pop and key_level == len(keys) - 2:
            dict_pop(d[level_key], keys[key_level + 1])
    elif isinstance(d, list):
        res = [dict_extract(item, keys, key_level=key_level, pop=pop) for item in d]
    else:
        res = None

    return res


def dict_clean(d: DictListType, keys: DictListType):
    """pop empty-value keys from dict or list recursively"""
    if not keys or not d:
        return
    keys = unify_key_to_list(keys)
    value = dict_get(d, keys)
    if not value:
        dict_extract(d, keys, pop=True)


def inner_dict_flatten(
    d: DictListType,
    value,
    new_key: str = None,
    level_keys: list = None,
    level: int = 0,
) -> DictListType:
    """recursive helper function for `dict_flatten()`"""
    level_key = level_keys[level]

    if isinstance(d, dict):
        if level == len(level_keys) - 1:
            d[new_key] = value
            return
        else:
            inner_dict_flatten(
                d[level_key],
                value=value,
                new_key=new_key,
                level_keys=level_keys,
                level=level + 1,
            )
    elif isinstance(d, list):
        for d_item, value_item in zip(d, value):
            inner_dict_flatten(
                d_item,
                value=value_item,
                new_key=new_key,
                level_keys=level_keys,
                level=level,
            )
    else:
        pass

    return


def dict_extract_and_flatten(d: DictListType, keys: KeysType, sep: str = "."):
    level_keys = deepcopy(keys)[:-1]
    new_key = sep.join(keys[-2:])
    value = dict_extract(d, keys, pop=True)
    inner_dict_flatten(d, value=value, new_key=new_key, level_keys=level_keys)


def dict_flatten(
    d: DictListType,
    keys: KeysType,
    sep: str = ".",
    in_replace: bool = True,
    expand_sub: bool = False,
) -> DictListType:
    """Flatten nested dict or list to single-level.
    Inputs:
    - d: input dict or list
    - keys: keys to flatten; if str, split by `sep`
    - sep: separator of joined keys
    - in_replace: if True, flatten original in-place and return; if False, do not modify original object, and return new one
    - expand_sub: if True, expand keys to multi sub_keys for dict; if False, keep as is
    """
    if not in_replace:
        xd = deepcopy(d)
    else:
        xd = d

    keys = unify_key_to_list(keys, sep=sep)
    sub_keys = None
    if expand_sub:
        value = dict_get(xd, keys)
        if isinstance(value, dict):
            sub_keys = [keys + [sub_key] for sub_key in value.keys()]

    if len(keys) < 2 and sub_keys is None:
        return xd

    if sub_keys:
        for sks in sub_keys:
            dict_extract_and_flatten(xd, keys=sks, sep=sep)
        dict_clean(xd, keys=keys)
    else:
        dict_extract_and_flatten(xd, keys=keys, sep=sep)

    return xd



""" Inspired by termcolor: https://github.com/termcolor/termcolor """

import re

from typing import Literal, Iterable, Union

FONT_TYPE = Literal[
    "bold",
    "dark",
    "underline",
    "blink",
    "reverse",
    "concealed",
]

BG_COLOR_TYPE = Literal[
    "bg_black",
    "bg_grey",
    "bg_red",
    "bg_green",
    "bg_yellow",
    "bg_blue",
    "bg_magenta",
    "bg_cyan",
    "bg_light_grey",
    "bg_dark_grey",
    "bg_light_red",
    "bg_light_green",
    "bg_light_yellow",
    "bg_light_blue",
    "bg_light_magenta",
    "bg_light_cyan",
    "bg_white",
]

COLOR_TYPE = Literal[
    "black",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_grey",
    "dark_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "white",
]

FONTS: dict[FONT_TYPE, int] = {
    "bold": 1,
    "dark": 2,
    "underline": 4,
    "blink": 5,
    "reverse": 7,
    "concealed": 8,
}

BG_COLORS: dict[BG_COLOR_TYPE, int] = {
    "bg_black": 40,
    "bg_grey": 40,
    "bg_red": 41,
    "bg_green": 42,
    "bg_yellow": 43,
    "bg_blue": 44,
    "bg_magenta": 45,
    "bg_cyan": 46,
    "bg_light_grey": 47,
    "bg_dark_grey": 100,
    "bg_light_red": 101,
    "bg_light_green": 102,
    "bg_light_yellow": 103,
    "bg_light_blue": 104,
    "bg_light_magenta": 105,
    "bg_light_cyan": 106,
    "bg_white": 107,
}

COLORS: dict[COLOR_TYPE, int] = {
    "black": 30,
    "grey": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "light_grey": 37,
    "dark_grey": 90,
    "light_red": 91,
    "light_green": 92,
    "light_yellow": 93,
    "light_blue": 94,
    "light_magenta": 95,
    "light_cyan": 96,
    "white": 97,
}


COLOR_SET = "\033[%sm%s"
COLOR_RESET = "\033[0m"
RE_COLORED = (
    r"(?P<colored_text>\033\[(?P<color_ints>\d+(\;\d+)*)m(?P<text>[^\033]*)\033\[0m)"
)


def color_text_with_ints(text: str, color_ints: list[int]) -> str:
    if color_ints:
        color_ints_str = ";".join(map(str, color_ints))
        res = f"\033[{color_ints_str}m{text}\033[0m"
    else:
        res = text
    return res


def colored(
    text: str,
    color: COLOR_TYPE = None,
    bg_color: BG_COLOR_TYPE = None,
    fonts: Union[FONT_TYPE, Iterable[FONT_TYPE]] = None,
) -> str:
    text = str(text)

    color_ints = []
    if color:
        color_ints.append(COLORS[color])
    if bg_color:
        color_ints.append(BG_COLORS[bg_color])
    if fonts:
        if isinstance(fonts, str):
            color_ints.append(FONTS[fonts])
        else:
            color_ints.extend([FONTS[font] for font in fonts])

    # handle nested colored text
    matches = re.finditer(RE_COLORED, text)
    if matches:
        res = ""
        prev_end = 0
        for match in matches:
            start = match.start()
            end = match.end()
            if start > prev_end:
                res += color_text_with_ints(text[prev_end:start], color_ints)
            res += match.group("colored_text")
            prev_end = end
        if prev_end < len(text):
            res += color_text_with_ints(text[prev_end:], color_ints)
    else:
        res = color_text_with_ints(text, color_ints)

    return res


def decolored(text: str) -> str:
    if not isinstance(text, str):
        return text
    matches = re.finditer(RE_COLORED, text)
    if matches:
        res = ""
        prev_end = 0
        for match in matches:
            start = match.start()
            end = match.end()
            if start > prev_end:
                res += text[prev_end:start]
            res += match.group("text")
            prev_end = end
        if prev_end < len(text):
            res += text[prev_end:]
    else:
        res = text

    return res



from typing import Literal
from .maths import chars_len


def chars_slice(
    text: str,
    beg: int = None,
    end: int = None,
    use_chars_len: bool = True,
    fill_char: str = " ",
    align: Literal["l", "r"] = None,
) -> str:
    """Slice string with beg and end, considering chars length."""
    if beg is None:
        beg = 0
    if end is None:
        end = len(text)

    if not use_chars_len:
        return text[beg:end]

    if beg < 0:
        beg = len(text) + beg
    if end < 0:
        end = len(text) + end + 1

    res_beg_idx = None
    res_end_idx = None
    fill_left = ""
    fill_right = ""
    res = ""
    beg_idx = 0
    end_idx = 0
    if not isinstance(text, str):
        if isinstance(text, (int, float)):
            align = "r"
        text = str(text)
    for ch in text:
        end_idx = beg_idx + chars_len(ch)
        if beg_idx >= beg and end_idx <= end:
            res += ch
            if res_beg_idx is None:
                res_beg_idx = beg_idx
                fill_left = fill_char * (beg_idx - beg)
        if end_idx >= end:
            if res_end_idx is None:
                res_end_idx = end_idx
                fill_right = fill_char * (end_idx - end)
            break
        beg_idx = end_idx
    if res_end_idx is None:
        res_end_idx = end_idx
        fill_right = fill_char * (end - end_idx)
    if not align:
        res = fill_left + res + fill_right
    elif align == "l":
        res = res + fill_left + fill_right
    else:
        res = fill_left + fill_right + res
    return res



def obj_param(obj, defaults=None, **kwargs):
    """Special case of obj_params, which only contains one kwarg."""
    if len(kwargs) == 0:
        return defaults

    if len(kwargs) == 1:
        key, val = next(iter(kwargs.items()))
        if val is not None:
            return val
        elif hasattr(obj, key):
            return getattr(obj, key)
        else:
            return defaults


def obj_params_dict(obj, defaults=None, **kwargs) -> dict:
    if len(kwargs) == 0:
        return defaults

    if len(kwargs) == 1:
        key, val = next(iter(kwargs.items()))
        if val is not None:
            return {key: val}
        elif hasattr(obj, key):
            return {key: getattr(obj, key)}
        else:
            return {key: defaults}

    # len(kwargs) > 1
    res = {}
    if defaults is None:
        defaults_dict = {}
    elif isinstance(defaults, (list, tuple)):
        defaults_dict = dict(zip(kwargs.keys(), defaults))
    elif isinstance(defaults, dict):
        defaults_dict = defaults

    for key, val in kwargs.items():
        if val is not None:
            res[key] = val
        elif hasattr(obj, key):
            res[key] = getattr(obj, key)
        else:
            res[key] = defaults_dict.get(key, None)
    return res


def obj_params_list(obj, defaults=None, **kwargs) -> list:
    return list(obj_params_dict(obj, defaults, **kwargs).values())


def obj_params_tuple(obj, defaults=None, **kwargs) -> tuple:
    return tuple(obj_params_dict(obj, defaults, **kwargs).values())


def obj_params(obj, defaults=None, **kwargs):
    """
    Resolve param with priority:
    1. kwargs
    2. obj attr
    3. defaults

    Usage:
    - `obj_params(obj, defaults="", model="model_name")`
    - `obj_params(obj, defaults=("", False), model="model_name", stream=True)`
    - `obj_params(obj, defaults={"model": "", "stream": False}, model="model_name", stream=True)`

    Inputs:
        Not allowed keys of kwargs: `obj`, `defaults`

    Return:
        0 kwarg : return `defaults`
        1 kwarg : return value
        N kwargs: return tuple

    Number of returned params is same with kwargs.
    """
    if len(kwargs) == 0:
        return defaults
    elif len(kwargs) == 1:
        return obj_param(obj, defaults, **kwargs)
    else:
        res_dict = obj_params_dict(obj, defaults, **kwargs)
        return tuple(res_dict.values())



from .types import StrsType, LIST_TYPES
from .maths import chars_len, is_str_float
from .fills import add_fills
from .colors import colored

from typing import Literal

HEAD_COLOR = "light_cyan"
CELL_COLOR = "light_blue"
SEPR_COLOR = "dark_grey"
VERT_COLOR = "dark_grey"
TOTAL_COLOR = "light_green"

SEPR = "-"
CSEPR = colored(SEPR, SEPR_COLOR)
VERT = "|"
CVERT = colored(VERT, VERT_COLOR)
WERT = f" | "


def is_listable(val) -> bool:
    return isinstance(val, LIST_TYPES)


def norm_any_to_str_list(val) -> list[str]:
    if is_listable(val):
        return [str(v) for v in val]
    else:
        return [str(val)]


def norm_any_to_type_list(val) -> list[str]:
    if is_listable(val):
        return [type(v).__name__ for v in val]
    else:
        return [type(val).__name__]


def add_bounds(line: str, is_colored: bool = False) -> str:
    if not is_colored:
        return f"{VERT} {line} {VERT}"
    else:
        return f"{CVERT} {line} {CVERT}"


def align_to_fill_side(align: str) -> Literal["left", "right", "both"]:
    if align[0].lower() == "l":
        return "right"
    elif align[0].lower() == "r":
        return "left"
    else:
        return "both"


def dict_to_table_str(
    d: dict,
    key_headers: StrsType = None,
    val_headers: StrsType = None,
    capitalize_headers: bool = True,
    aligns: StrsType = None,
    default_align: Literal["left", "right"] = "right",
    sum_at_tail: bool = False,
    is_colored: bool = False,
) -> str:
    if not d:
        return ""

    if not key_headers or not val_headers:
        k1, v1 = next(iter(d.items()))
        if not key_headers:
            key_headers = norm_any_to_str_list(k1)
        if not val_headers:
            val_headers = norm_any_to_str_list(v1)

    table_headers: list[str] = key_headers + val_headers
    if capitalize_headers:
        table_headers = [h.capitalize() for h in table_headers]

    table_rows: list[list[str]] = []
    for key, val in d.items():
        key_strs = norm_any_to_str_list(key)
        val_strs = norm_any_to_str_list(val)
        row = key_strs + val_strs
        table_rows.append(row)

    cols = len(table_headers)

    # add total row
    is_sum_at_tail = sum_at_tail and cols > 1
    if is_sum_at_tail:
        sum_row: list[str] = []
        for i in range(cols):
            is_sum_valid = True
            col_sum = 0
            for row in table_rows:
                cell = row[i]
                if not cell or cell == "-":
                    continue
                elif isinstance(cell, (int, float)):
                    col_sum += cell
                elif cell.isdigit():
                    col_sum += int(cell)
                elif is_str_float(cell):
                    col_sum += float(cell)
                else:
                    is_sum_valid = False
                    break
            if is_sum_valid and col_sum != 0:
                sum_row.append(str(col_sum))
            else:
                sum_row.append("")
        if any(v for v in sum_row):
            is_sum_at_tail = True
            # prepend "Total" label to numeric columns
            for i in range(1, cols):
                if sum_row[i]:
                    sum_row[i - 1] = "Total"
                    break
            table_rows.append(sum_row)
        else:
            is_sum_at_tail = False

    table_headers_rows = [table_headers] + table_rows
    col_widths = [
        max(chars_len(row[i]) if i < len(row) else 0 for row in table_headers_rows)
        for i in range(cols)
    ]

    sep_lines = [SEPR * col_widths[i] for i in range(cols)]
    wert = WERT

    if is_colored:
        table_headers = [colored(h, HEAD_COLOR) for h in table_headers]
        if is_sum_at_tail:
            colored_sum_row = [
                colored(cell, TOTAL_COLOR) if cell else cell for cell in table_rows[-1]
            ]
            colored_non_sum_rows = [
                [colored(cell, CELL_COLOR) for cell in row] for row in table_rows[:-1]
            ]
            table_rows = colored_non_sum_rows + [colored_sum_row]
        else:
            table_rows = [
                [colored(cell, CELL_COLOR) for cell in row] for row in table_rows
            ]
        sep_lines = [colored(s, SEPR_COLOR) for s in sep_lines]
        wert = colored(WERT, VERT_COLOR)

    if not aligns:
        aligns = [default_align] * cols
    if len(aligns) < cols:
        aligns += [default_align] * (cols - len(aligns))

    header_line_str = wert.join(
        add_fills(
            text=table_headers[i],
            filler=" ",
            fill_side=align_to_fill_side(aligns[i]),
            is_text_colored=is_colored,
            total_width=col_widths[i],
        )
        for i in range(cols)
    )
    header_line_str = add_bounds(header_line_str, is_colored=is_colored)

    sep_line_str = wert.join(sep_lines)
    sep_line_str = add_bounds(sep_line_str, is_colored=is_colored)

    rows_lines = []
    for row in table_rows:
        row_line = wert.join(
            add_fills(
                text=row[i],
                filler=" ",
                fill_side=align_to_fill_side(aligns[i]),
                is_text_colored=is_colored,
                total_width=col_widths[i],
            )
            for i in range(cols)
        )
        row_line = add_bounds(row_line, is_colored=is_colored)
        rows_lines.append(row_line)

    if is_sum_at_tail:
        row_lines_str = (
            "\n".join(rows_lines[:-1]) + f"\n{sep_line_str}\n{rows_lines[-1]}"
        )
    else:
        row_lines_str = "\n".join(rows_lines)

    table_str = f"{header_line_str}\n{sep_line_str}\n{row_lines_str}"
    return table_str
